package cn.ecpark.service.usergw.config;

import java.io.InputStream;
import java.lang.reflect.Method;
import java.net.URI;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.dubbo.config.ReferenceConfig;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.SpringApplication;
import org.springframework.cloud.gateway.filter.FilterDefinition;
import org.springframework.cloud.gateway.handler.predicate.PredicateDefinition;
import org.springframework.cloud.gateway.route.RouteDefinition;
import org.springframework.cloud.gateway.route.RouteDefinitionLocator;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.stereotype.Component;
import org.springframework.stereotype.Controller;
import org.yaml.snakeyaml.Yaml;

import cn.ecpark.service.usergw.biz.filters.bean.GenericServicePool;
import cn.ecpark.service.usergw.biz.filters.factory.DubboFilterFactory;
import cn.ecpark.service.usergw.impl.http.Http2Dubbo;
import cn.ecpark.service.usergw.utils.Convert;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Flux;

/**
 * ConfigGateway
 */
@Component
@Controller
@Slf4j
public class ConfigGateway implements RouteDefinitionLocator {

  // List<MediaType> mediaTypes = new ArrayList<>();

  List<FilterDefinition> defaultFilters;

  @Autowired
  private ApplicationContext applicationContext;

  @Bean
  public GenericServicePool genericServicePool() {
    return new GenericServicePool();
  }

  @Bean
	public DubboFilterFactory dubboFilterFactory() {
		return new DubboFilterFactory();
	}

  @Autowired
  private Http2Dubbo http2Dubbo;

  @Override
  @SuppressWarnings("unchecked")
  public Flux<RouteDefinition> getRouteDefinitions() {
    // WebFluxConfigurationSupport a;

    Object inputStream = null;
    String[] gatewayConfigPathList = { "gateway.yml", "gateway.yaml", "Gateway.yml", "Gateway.yaml" };
    for (String gatewayConfigPath : gatewayConfigPathList) {
      inputStream = this.getClass().getClassLoader().getResourceAsStream(gatewayConfigPath);
      if (inputStream != null) {
        break;
      }
    }

    if (inputStream != null) {
      Map<String, Object> configYaml = new Yaml().load((InputStream) inputStream);
      if (configYaml != null) {

        // streaming-media-types 这个先不管
        // NettyWriteResponseFilter nwrf =
        // applicationContext.getBean(NettyWriteResponseFilter.class);
        // try {
        // Field field = nwrf.getClass().getDeclaredField("streamingMediaTypes");
        // field.setAccessible(true);
        // List<MediaType> mts = new ArrayList<MediaType>();
        // mts.add(MediaType.TEXT_EVENT_STREAM);
        // mts.add(MediaType.APPLICATION_STREAM_JSON);
        // field.set(nwrf, mts);
        // } catch (Exception e) {
        // log.error(e.toString());
        // }

        Map<String, Object> defaultYaml = (Map<String, Object>) configYaml.get("default");
        Map<String, Object> dubboYaml = (Map<String, Object>) configYaml.get("dubbo");

        List<RouteDefinition> routeList = new ArrayList<RouteDefinition>();

        if (defaultYaml != null && !defaultYaml.isEmpty()) {
          defaultFilters = this.getDefaultFilter(defaultYaml);
          this.configDefault(routeList, defaultYaml);
        }

        if (dubboYaml != null && !dubboYaml.isEmpty()) {
          this.configHttp2Dubbo(routeList, dubboYaml);
        }

        if (!routeList.isEmpty()) {
          return Flux.fromIterable(routeList);
        }

      }
    }

    return Flux.empty();
    // return Flux.fromIterable(it);
  }

  @SuppressWarnings("unchecked")
  private List<FilterDefinition> getDefaultFilter(Map<String, Object> defaultYaml) {
    List<FilterDefinition> filters = new ArrayList<>();

    // default-filters: 下的相关设置
    Object unknownDefaultFilters = defaultYaml.get("default-filters");
    if (unknownDefaultFilters != null) {
      List<String> defaultFiltersYaml = (ArrayList<String>) unknownDefaultFilters;
      for (String filterString : defaultFiltersYaml) {
        filters.add(new FilterDefinition(filterString));
      }
    }

    return filters;
  }

  @SuppressWarnings("unchecked")
  private void configDefault(List<RouteDefinition> routeList, Map<String, Object> defaultYaml) {
    if (defaultYaml != null) {

      Object unknownRoutes = defaultYaml.get("routes");
      if (unknownRoutes != null) {
        List<LinkedHashMap<String, List<String>>> routes = (ArrayList<LinkedHashMap<String, List<String>>>) unknownRoutes;
        for (LinkedHashMap<String, List<String>> iter : routes) {

          List<FilterDefinition> filters = new ArrayList<>();
          List<PredicateDefinition> predicates = new ArrayList<>();
          RouteDefinition rd = new RouteDefinition();



          // 设置基础属性
          this.ParseAndSetBase(rd, iter);

          // predicates: 下的相关属性
          this.ParseAndAddPredicates(predicates, iter, "predicates");

          // filters: 下的相关属性
          this.ParseAndAddFilters(filters, iter, "filters");

          rd.setPredicates(predicates);
          rd.setFilters(filters);

          routeList.add(rd);
        }
      }
    }
  }

  private void ParseAndSetBase(RouteDefinition rd, LinkedHashMap<String, List<String>> iter) {
    // 设置id
    Object id = iter.get("id");
    if (id != null) {
      rd.setId((String) id);
    }

    // 设置uri
    Object uri = iter.get("uri");
    if (uri != null) {
      rd.setUri(URI.create((String) uri));
    }

    // 设置uri
    Object order = iter.get("order");
    if (order != null) {
      rd.setOrder((int) order);
    }
  }

  private void ParseAndAddPredicates(List<PredicateDefinition> predicates, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> predicatesYaml = iter.get(yamlField);
    if (predicatesYaml != null) {
      for (String predicateString : predicatesYaml) {
        predicates.add(new PredicateDefinition(predicateString));
      }
    }
  }

  private void ParseAndAddFilters(List<FilterDefinition> filters, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> filtersYaml = iter.get(yamlField);
    if (filtersYaml != null) {
      filters.addAll(defaultFilters);
      for (String filterString : filtersYaml) {
        filters.add(new FilterDefinition(filterString));
      }
    }
  }

  private void configHttp2Dubbo(List<RouteDefinition> routeList, Map<String, Object> configDubbo) {

    try {

      Object unknownRoutes = configDubbo.get("routes");
      if (unknownRoutes != null) {
        GenericServicePool pool = applicationContext.getBean(GenericServicePool.class);

        List<LinkedHashMap<String, List<String>>> routes = (ArrayList<LinkedHashMap<String, List<String>>>) unknownRoutes;
        for (LinkedHashMap<String, List<String>> iter : routes) {

          List<FilterDefinition> filters = new ArrayList<>();
          List<PredicateDefinition> predicates = new ArrayList<>();
          RouteDefinition rd = new RouteDefinition();

          this.ParseAndSetBase(rd, iter);

          // 设置基础属性
          this.ParseAndSetBase(rd, iter);

          // predicates: 下的相关属性
          this.ParseAndAddPredicates(predicates, iter, "predicates");

          // filters: 下的相关属性
          this.ParseAndAddFilters(filters, iter, "filters");

        }

      }

      // RequestMappingHandlerMapping requestMapping = (RequestMappingHandlerMapping)
      // applicationContext
      // .getBean("requestMappingHandlerMapping");

      // requestMapping.registerMapping(RequestMappingInfo.paths("/test/xixi").methods(RequestMethod.POST).build(),
      // http2Dubbo, Http2Dubbo.class.getDeclaredMethod("H2DTest"));

      // 引用远程服务
      // try {
      // ReferenceConfig<GenericService> reference = new
      // ReferenceConfig<GenericService>(); // 该实例很重量，里面封装了所有与注册中心及服务提供方连接，请缓存

      // reference.setInterface("com.xxx.XxxService"); // 弱类型接口名
      // reference.setVersion("2.0.0");
      // reference.setGeneric(true); // 声明为泛化接口
      // GenericService gs = reference.get();

    } catch (Exception e) {
      // TODO: 任何错误都直接终止退出
      log.error(e.toString());
      SpringApplication.exit(applicationContext);
    }

    // gs.$invoke(method, parameterTypes, args)

  }

  private static void callMethod(ReferenceConfig ref, String name, String params) throws Exception {
    Method method = ref.getClass().getMethod("set" + Convert.firstUpperCase(name));
    method.invoke(ref, method.getParameterTypes(), params);
  }
}